#define _GNU_SOURCE
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <dirent.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <limits.h>

#include "uhumenu.h"

char *plugin_name = "desktop";

static char *file_content;
static int file_size;

static int path_remove;

static int parse_line (struct item *item, struct section **sp, char *s)
{
	char *lang, *value;
	char *end = s + strlen(s);
	while (end > s && (*(end-1) == ' ' || *(end-1) == '\t')) *(--end) = '\0';
	while (*s == ' ' || *s == '\t') s++;
	if (*s == '\0' || *s == '#') return 0;
	if (*s == '[' && *(end-1) == ']') {
		*(--end) = '\0';
		*sp = item_add_section(item, s+1);
		return 0;
	}
	value = strchr(s, '=');
	if (value == NULL) return -1;
	end = value;
	*end = '\0';
	value++;
	while (*value == ' ' || *value == '\t') value++;
	while (end > s && (*(end-1) == ' ' || *(end-1) == '\t')) *(--end) = '\0';
	if (*s == '\0') return -1;
	lang = strchr(s, '[');
	if (lang != NULL && *(end-1) == ']') {
		*lang++ = '\0';
		*(--end) = '\0';
	} else {
		lang = "";
	}
	if (*sp == NULL) {
		fprintf(stderr, "section == NULL !!!\n");
		return -1;
	}
	section_add_value(*sp, s, lang, value);
	return 0;
}

static int parse_item (struct item *item)
{
	char *p, *s;
	struct section *section = NULL;

	item->sections = NULL;
	p = file_content;
	while (p < file_content + file_size) {
		s = strchr(p, '\n');
		if (s == NULL) {
			s = p + strlen(p);
		} else {
			*s++ = '\0';
		}
		parse_line(item, &section, p);
		p = s;
	}

	s = get_value(item, NULL, "Terminal", NULL);
	item->terminal = (s != NULL && (!strcmp(s, "true") || !strcmp(s, "1")));

	item->wmlist = get_value(item, NULL, "WindowManagers", NULL);

	return 0;
}

static int file_to_item (const char *file, struct item *item)
{
	struct stat st;
	int i;
	char buf[PATH_MAX];
	const char *filename;
	int fd;

	i = stat(file, &st);
	if (i < 0) {
		fprintf(stderr, "cannot stat %s: %s\n", file, strerror(errno));
		return -1;
	}
	if (S_ISREG(st.st_mode)) {
		item->issubmenu = 0;
		filename = file;
	} else if (S_ISDIR(st.st_mode)) {
		item->issubmenu = 1;
		sprintf(buf, "%s/.directory", file);
		filename = buf;
	} else {
		fprintf(stderr, "%s is not a regular file or directory\n", file);
		return -1;
	}

	fd = open(filename, O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "cannot open %s: %s\n", filename, strerror(errno));
		return -1;
	}

	if (fstat(fd, &st) < 0) {
		fprintf(stderr, "fstat: %s\n", strerror(errno));
		close(fd);
		return -1;
	}

	item->file_name = strdup(file + path_remove);
	file_size = st.st_size;
	file_content = malloc(file_size + 1);

	if (read(fd, file_content, file_size) != file_size) {
		fprintf(stderr, "read failed\n");
		close(fd);
		return -1;
	}
	file_content[file_size] = '\0';

	close(fd);

	return 0;
}

static struct dirent *readdir_nobackup (DIR *d)
{
	struct dirent *de;
	while ((de = readdir(d)) != NULL) {
		if (de->d_name[0] == '.') continue;
		if (strstr(de->d_name, ".dpkg") != NULL) continue;
		if (strstr(de->d_name, ".rpm") != NULL) continue;
		if (strstr(de->d_name, ".orig") != NULL) continue;
		if (strstr(de->d_name, ".bak") != NULL) continue;
		if (strstr(de->d_name, "~") != NULL) continue;
		return de;
	}
	return NULL;
}

static struct menu *build_menu (const char *dir)
{
	DIR *d;
	struct dirent *de;
	char buf[PATH_MAX];
	struct menu *menu;
	int i;

	d = opendir(dir);
	if (d == NULL) {
		fprintf(stderr, "opendir failed: %s\n", strerror(errno));
		return NULL;
	}
	menu = malloc(sizeof(struct menu));
	menu->nr = 0;
	while ((de = readdir_nobackup(d)) != NULL) menu->nr++;
	menu->items = malloc(menu->nr * sizeof(struct item));
	rewinddir(d);
	i = 0;
	while (i < menu->nr && (de = readdir_nobackup(d)) != NULL) {
		sprintf(buf, "%s/%s", dir, de->d_name);
		if (file_to_item(buf, &menu->items[i]) < 0) continue;
		if (parse_item(&menu->items[i]) < 0) continue;
		free(file_content);
		if (menu->items[i].issubmenu) {
			if ((menu->items[i].submenu = build_menu(buf)) == NULL) {
				continue;
			}
		}
		i++;
	}
	menu->nr = i;
	closedir(d);
	if (i == 0) {
		free(menu->items);
		free(menu);
		return NULL;
	}
	return menu;
}

struct menu *read_menu (const char *path)
{
	struct menu *menu;

	path_remove = strlen(path);

	menu = malloc(sizeof(struct menu));
	menu->nr = 1;
	menu->items = malloc(sizeof(struct item));
	if (file_to_item(path, &menu->items[0]) < 0) return NULL;
	if (!menu->items[0].issubmenu) {
		fprintf(stderr, "%s is not a directory\n", path);
		return NULL;
	}
	if (parse_item(&menu->items[0]) < 0) return NULL;
	free(file_content);
	if ((menu->items[0].submenu = build_menu(path)) == NULL) menu->nr = 0;

	return menu;
}
