#define _GNU_SOURCE
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <locale.h>
#include <langinfo.h>
#include <iconv.h>
#include <dlfcn.h>
#include <dirent.h>
#include <getopt.h>
#include <errno.h>
#include <sys/stat.h>

#define UHU_MENU_CORE_ENGINE
#include "uhumenu.h"

#define MENU_SRC_DIR "/usr/lib/menu/SOURCE"
#define MENU_DST_DIR "/usr/lib/menu"
#define PLUGIN_DIR "/usr/lib/uhumenu"
#define INPUT_PLUGIN "desktop"

const char *menu_src_dir = MENU_SRC_DIR;
const char *menu_dst_dir = MENU_DST_DIR;
const char *plugin_dir = PLUGIN_DIR;
const char *input_plugin = INPUT_PLUGIN;
char input_plugin_dir[PATH_MAX];
char output_plugin_dir[PATH_MAX];

char buf[PATH_MAX];

char *wm_only = NULL;
char *wm_skip = NULL;

struct language *current_language;

#define PLUGINS_MAX 64
struct plugin {
	char name[FILENAME_MAX + 1];
	int ok;
	void *handle;
	int (*write_menu) (const struct menu *menu, const char *path);
	int (*write_menu_lang) (const struct menu *menu, const char *path, const struct language *language);
};

struct plugin plugins[PLUGINS_MAX];
int pluginnr = 0;

int string_to_int (const char *s, int err)
{
	int val;
	char dummy;
	if (sscanf(s, "%d%c", &val, &dummy) != 1) return err;
	return val;
}

int comma_substring (const char *haystack, const char *needle)
{
	int retval;
	char *h2 = malloc(strlen(haystack) + 3);
	char *n2 = malloc(strlen(needle) + 3);
	sprintf(h2, ",%s,", haystack);
	sprintf(n2, ",%s,", needle);
	retval = (strstr(h2, n2) != NULL);
	free(n2);
	free(h2);
	return retval;
}

int plugin_sort_fn (const void *v1, const void *v2)
{
	const struct plugin *p1 = v1;
	const struct plugin *p2 = v2;
	return strcmp(p1->name, p2->name);
}

/*
 *  qsort_menu + menu_sort_fn
 *  Sorts the menu according to current_lang. Invalid entries are put to the
 *  end of the list. "nr" is set to the number of valid entries, as opposed to
 *  nr_real which counts the invalid entries too.
 */
int menu_sort_fn (const void *v1, const void *v2)
{
	const struct item *i1 = v1;
	const struct item *i2 = v2;
	const struct key *k1, *k2;
	if (i1->valid != i2->valid) return i2->valid - i1->valid;
	if (i1->issubmenu != i2->issubmenu) return i2->issubmenu - i1->issubmenu;
	if (i1->major != i2->major) return i2->major - i1->major;
	if (i1->minor != i2->minor) return i2->minor - i1->minor;
	k1 = item_find_key(i1, NULL, "Name");
	k2 = item_find_key(i2, NULL, "Name");
	if (k1 == NULL && k2 == NULL) return 0;
	if (k1 == NULL) return -1;
	if (k2 == NULL) return 1;
	if (k1->current_lang == NULL && k2->current_lang == NULL) return 0;
	if (k1->current_lang == NULL) return -1;
	if (k2->current_lang == NULL) return 1;
	if (current_language->locale_utf8 != NULL) {
		return strcoll_l(k1->current_lang->value, k2->current_lang->value, current_language->locale_utf8);
	} else if (current_language->locale != NULL) {
		return strcoll_l(k1->current_lang->value_8bit, k2->current_lang->value_8bit, current_language->locale);
	} else {
		return strcmp(k1->current_lang->value, k2->current_lang->value);
	}
}

void qsort_menu (struct menu *menu)
{
	int i;
	qsort(menu->items, menu->nr_real, sizeof(struct item), menu_sort_fn);
	for (i = 0; i < menu->nr_real && menu->items[i].valid; i++) {
		if (menu->items[i].issubmenu) {
			qsort_menu(menu->items[i].submenu);
		}
	}
	menu->nr = i;
}

/*
 *  empty_is_invalid
 *  Marks empty submenus (those which do not have valid and nonempty descendants)
 *  as invalid.
 */
int empty_is_invalid (struct menu *menu)
{
	int retval = -1;
	int i;
	for (i = 0; i < menu->nr_real; i++) {
		if (!menu->items[i].valid) continue;
		if (menu->items[i].issubmenu && empty_is_invalid(menu->items[i].submenu) < 0) {
			menu->items[i].valid = 0;
		} else {
			retval = 0;
		}
	}
	return retval;
}

/*
 *  prepare_for_wm
 *  Prepares the menu for a particular window manager. Marks entries that
 *  should be skipped for this wm as invalid.
 */
int prepare_for_wm (struct menu *menu, const char *wm)
{
	int i;
	for (i = 0; i < menu->nr_real; i++) {
		menu->items[i].valid_wm = 1;
		if (menu->items[i].wmlist != NULL &&
		  !comma_substring(menu->items[i].wmlist, wm)) {
			menu->items[i].valid_wm = 0;
		}
		menu->items[i].valid = menu->items[i].valid_wm && menu->items[i].valid_lang;
		if (menu->items[i].issubmenu) {
			prepare_for_wm(menu->items[i].submenu, wm);
		}
	}
	return 0;
}

/*
 *  prepare_for_lang
 *  Prepares the menu for a particular language. Sets the current_lang pointer
 *  in every key of every item for faster access. Converts major and minor
 *  entries from string to int. Does not yet sort the entries.
 */
int prepare_for_lang (struct menu *menu, const char *lang)
{
	int i;
	struct section *s;
	struct key *k;
	for (i = 0; i < menu->nr_real; i++) {
		s=menu->items[i].sections;
		while (s != NULL) {
			k = s->keys;
			while (k != NULL) {
				k->current_lang = key_find_lang(k, lang);
				k = k->next;
			}
			s = s->next;
		}
		menu->items[i].valid_lang = 1;
		if (menu->items[i].key_name == NULL || menu->items[i].key_name->current_lang == NULL) {
			menu->items[i].valid_lang = 0;
		}
		if (menu->items[i].key_major == NULL ||
		  menu->items[i].key_major->current_lang == NULL) {
			menu->items[i].major = 0;
		} else {
			menu->items[i].major = string_to_int(menu->items[i].key_major->current_lang->value, 0);
		}
		if (menu->items[i].key_minor == NULL ||
		  menu->items[i].key_minor->current_lang == NULL) {
			menu->items[i].minor = 0;
		} else {
			menu->items[i].minor = string_to_int(menu->items[i].key_minor->current_lang->value, 0);
		}
		menu->items[i].valid = menu->items[i].valid_wm && menu->items[i].valid_lang;
		if (menu->items[i].issubmenu) {
			prepare_for_lang(menu->items[i].submenu, lang);
		}
	}
	return 0;
}

/*
 *  fix_menu
 *  Called only once after the input plugin has built up the menu tree and
 *  before the input plugins dump it. Fills out some fields which might yet
 *  be uninitialized but may be needed by this core engine or an output plugin.
 */
int fix_menu (struct menu *menu)
{
	int i;
	struct section *s;
	struct key *k;
	struct lang *l;
	menu->nr_real = menu->nr;
	for (i = 0; i < menu->nr; i++) {
		menu->items[i].key_name  = item_find_key(&menu->items[i], "", "Name");
		menu->items[i].key_major = item_find_key(&menu->items[i], "", "Major");
		menu->items[i].key_minor = item_find_key(&menu->items[i], "", "Minor");
		s = menu->items[i].sections;
		while (s != NULL) {
			k = s->keys;
			while (k != NULL) {
				l = k->langs;
				while (l != NULL) {
					if (l->language->charset != NULL) {
						if (l->value == NULL) {
							l->value = convert(l->language->charset->to_utf8, l->value_8bit);
						} else {
							l->value_8bit = convert(l->language->charset->from_utf8, l->value);
						}
					}
					l = l->next;
				}
				k = k->next;
			}
			s = s->next;
		}
		if (menu->items[i].issubmenu) {
			fix_menu(menu->items[i].submenu);
		}
	}
	return 0;
}

void list ()
{
	int i;
	for (i = 0; i < pluginnr; i++) {
		if (!plugins[i].ok) continue;
		printf("%-15s (%s)\n",
		  plugins[i].name,
		  plugins[i].write_menu_lang != NULL ? "separate config for each language"
		    : "one config for all languages");
	}
}

void help ()
{
	printf(
	  "Usage: uhumenu [options...]\n"
	  "\n"
	  "  -s, --source-dir       Source dir of menu (default: " MENU_SRC_DIR ")\n"
	  "  -d, --destination-dir  Destination dir of menu (default: " MENU_DST_DIR ")\n"
	  "  -p, --plugin-dir       Plugin dir (default: " PLUGIN_DIR ")\n"
	  "  -i, --input-plugin     Input plugin (default: " INPUT_PLUGIN ")\n"
	  "  -l, --list-plugins     List available window manager plugins\n"
	  "      --wm-only          Process only these window managers (comma separated)\n"
	  "      --wm-skip          Skip these window managers (comma separated)\n"
	  "  -h, --help             Guess what :-)\n"
	);
}

int main (int argc, char *argv[])
{
	struct menu *menu;
	void *ip;
	struct menu *(*read_menu) (const char *path);
	const char *err1, *err2;
	DIR *d;
	struct dirent *de;
	int c;
	int l;
	int i;
	int opt_list = 0;

	static const struct option long_options[] = {
		{ "source-dir",      1, NULL, 's' },
		{ "src-dir",         1, NULL, 's' },
		{ "destination-dir", 1, NULL, 'd' },
		{ "dest-dir",        1, NULL, 'd' },
		{ "dst-dir",         1, NULL, 'd' },
		{ "plugin-dir",      1, NULL, 'p' },
		{ "input-plugin",    1, NULL, 'i' },
		{ "list-plugins",    0, NULL, 'l' },
		{ "wm-only",         1, NULL, 256 },
		{ "wm-skip",         1, NULL, 257 },
		{ "help",            0, NULL, 'h' },
		{ NULL,              0, NULL, 0   }
	};

	/* Initialization */

	setlocale(LC_ALL, "");

	find_or_create_charset("UTF-8");
	find_or_create_language("");

	while ((c = getopt_long(argc, argv, "s:d:p:i:lh", long_options, NULL)) != -1) {
		switch (c) {
		  case 's':
			menu_src_dir = optarg;
			break;
		  case 'd':
			menu_dst_dir = optarg;
			break;
		  case 'p':
			plugin_dir = optarg;
			break;
		  case 'i':
			input_plugin = optarg;
			break;
		  case 'l':
			opt_list = 1;
			break;
		  case 256:
		  case 257:
			if (wm_only != NULL || wm_skip != NULL) {
				fprintf(stderr, "Window manager list already specified\n");
				exit(1);
			}
			if (c == 256) wm_only = optarg;
			else wm_skip = optarg;
			break;
		  case 'h':
			help();
			exit(0);
		  case '?':
		  case ':':
			exit(1);
		}
	}
	sprintf(input_plugin_dir, "%s/plugins/input", plugin_dir);
	sprintf(output_plugin_dir, "%s/plugins/output", plugin_dir);

	/* Scan output plugin candidates */

	d = opendir(output_plugin_dir);
	if (d == NULL) {
		fprintf(stderr, "Cannot open plugin dir %s: %s\n", output_plugin_dir, strerror(errno));
		exit(1);
	}
	while ((de = readdir(d)) != NULL) {
		if (!strcmp(de->d_name, "." )) continue;
		if (!strcmp(de->d_name, "..")) continue;
		l = strlen(de->d_name);
		if (l < 4
		  || strcmp(de->d_name + l - 3, ".so")) {
			fprintf(stderr, "Skipping non-plugin %s/%s\n", output_plugin_dir, de->d_name);
			continue;
		}
		if (pluginnr == PLUGINS_MAX) {
			fprintf(stderr, "Too many plugins found\n");
			exit(1);
		}
		sprintf(buf, "%.*s", l - 3, de->d_name);
		if (wm_only != NULL && !comma_substring(wm_only, buf)) continue;
		if (wm_skip != NULL &&  comma_substring(wm_skip, buf)) continue;
		strcpy(plugins[pluginnr].name, buf);
		plugins[pluginnr].ok = 1;
		pluginnr++;
	}
	closedir(d);
	qsort(plugins, pluginnr, sizeof(struct plugin), plugin_sort_fn);

	/* Open output plugins */

	for (i = 0; i < pluginnr; i++) {
		sprintf(buf, "%s/%s.so", output_plugin_dir, plugins[i].name);
		plugins[i].handle = dlopen(buf, RTLD_NOW);
		err1 = dlerror();
		if (err1 != NULL) {
			fprintf(stderr, "dlopen %s: %s\n", buf, err1);
			continue;
		}
		plugins[i].write_menu = dlsym(plugins[i].handle, "write_menu");
		err1 = dlerror();
		plugins[i].write_menu_lang = dlsym(plugins[i].handle, "write_menu_lang");
		err2 = dlerror();
		if ((err1 == NULL && err2 == NULL) || (err1 != NULL && err2 != NULL)) {
			plugins[i].ok = 0;
			fprintf(stderr, "Ignoring faulty output plugin %s\n", plugins[i].name);
			dlclose(plugins[i].handle);
			continue;
		}
	}

	if (opt_list) {
		list();
		exit(0);
	}

	/* Open input plugin */

	sprintf(buf, "%s/%s.so", input_plugin_dir, input_plugin);
	ip = dlopen(buf, RTLD_NOW);
	err1 = dlerror();
	if (err1 != NULL) {
		fprintf(stderr, "dlopen %s: %s\n", buf, err1);
		exit(1);
	}
	read_menu = dlsym(ip, "read_menu");
	err1 = dlerror();
	if (err1 != NULL) {
		fprintf(stderr, "Faulty input plugin!\n");
		exit(1);
	}

	/* Read the menu */

	menu = (*read_menu)(menu_src_dir);

	if (!menu) {
		fprintf(stderr, "Could not read menu!\n");
		exit(1);
	}

	/* Fixing the menu (filling some fields etc.) */

	fix_menu(menu);

	/* Process all-in-one output plugins */

	current_language = find_or_create_language("");
	prepare_for_lang(menu, "");

	for (i = 0; i < pluginnr; i++) {
		if (!plugins[i].ok) continue;
		if (plugins[i].write_menu == NULL) continue;

		prepare_for_wm(menu, plugins[i].name);
		empty_is_invalid(menu);
		qsort_menu(menu);

		sprintf(buf, "%s/%s", menu_dst_dir, plugins[i].name);
		mkdir(buf, 0777);
		(*plugins[i].write_menu)(menu, buf);
	}

	/* Process once-for-each-language plugins */

	current_language = first_language;
	while (current_language != NULL) {

/* XXX */
if (strcmp(current_language->name, "") && strcmp(current_language->name, "hu")) {
	current_language = current_language->next;
	continue;
}
/* XXX */

		prepare_for_lang(menu, current_language->name);
		for (i = 0; i < pluginnr; i++) {
			if (!plugins[i].ok) continue;
			if (plugins[i].write_menu_lang == NULL) continue;

			prepare_for_wm(menu, plugins[i].name);
			empty_is_invalid(menu);
			qsort_menu(menu);

			sprintf(buf, "%s/%s", menu_dst_dir, plugins[i].name);
			mkdir(buf, 0777);
			(*plugins[i].write_menu_lang)(menu, buf, current_language);
		}
		current_language = current_language->next;
	}

	/* Close output plugins and input plugin */

	for (i = 0; i < pluginnr; i++) {
		if (plugins[i].ok) dlclose(plugins[i].handle);
	}
	dlclose(ip);

	/* Yippie! Survived without a segfault :-) */

	return 0;
}
